@@ -19,7 +19,7 @@
 #endif
 
 #define OP_ZVAL_DTOR(op) do { \
-	(op).u.constant.is_ref = 0; \
+	Z_UNSET_ISREF((op).u.constant); \
 	zval_dtor(&(op).u.constant); \
 } while(0)
 xc_compile_result_t *xc_compile_result_init(xc_compile_result_t *cr, /* {{{ */
@@ -143,6 +143,9 @@ int xc_undo_pass_two(zend_op_array *op_array TSRMLS_DC) /* {{{ */
 	while (opline < end) {
 #ifdef ZEND_ENGINE_2_1
 		switch (opline->opcode) {
+#ifdef ZEND_GOTO
+			case ZEND_GOTO:
+#endif
 			case ZEND_JMP:
 				opline->op1.u.opline_num = opline->op1.u.jmp_addr - op_array->opcodes;
 				assert(opline->op1.u.opline_num < op_array->last);
@@ -151,6 +154,9 @@ int xc_undo_pass_two(zend_op_array *op_array TSRMLS_DC) /* {{{ */
 			case ZEND_JMPNZ:
 			case ZEND_JMPZ_EX:
 			case ZEND_JMPNZ_EX:
+#ifdef ZEND_JMP_SET
+			case ZEND_JMP_SET:
+#endif
 				opline->op2.u.opline_num = opline->op2.u.jmp_addr - op_array->opcodes;
 				assert(opline->op2.u.opline_num < op_array->last);
 				break;
@@ -180,15 +186,19 @@ int xc_redo_pass_two(zend_op_array *op_array TSRMLS_DC) /* {{{ */
 	end = opline + op_array->last;
 	while (opline < end) {
 		if (opline->op1.op_type == IS_CONST) {
-			opline->op1.u.constant.is_ref = 1;
-			opline->op1.u.constant.refcount = 2; /* Make sure is_ref won't be reset */
+			Z_SET_ISREF(opline->op1.u.constant);
+			Z_SET_REFCOUNT(opline->op1.u.constant, 2); /* Make sure is_ref won't be reset */
+
 		}
 		if (opline->op2.op_type == IS_CONST) {
-			opline->op2.u.constant.is_ref = 1;
-			opline->op2.u.constant.refcount = 2;
+			Z_SET_ISREF(opline->op2.u.constant);
+			Z_SET_REFCOUNT(opline->op2.u.constant, 2);
 		}
 #ifdef ZEND_ENGINE_2_1
 		switch (opline->opcode) {
+#ifdef ZEND_GOTO
+			case ZEND_GOTO:
+#endif
 			case ZEND_JMP:
 				assert(opline->op1.u.opline_num < op_array->last);
 				opline->op1.u.jmp_addr = op_array->opcodes + opline->op1.u.opline_num;
@@ -197,6 +207,9 @@ int xc_redo_pass_two(zend_op_array *op_array TSRMLS_DC) /* {{{ */
 			case ZEND_JMPNZ:
 			case ZEND_JMPZ_EX:
 			case ZEND_JMPNZ_EX:
+#ifdef ZEND_JMP_SET
+			case ZEND_JMP_SET:
+#endif
 				assert(opline->op2.u.opline_num < op_array->last);
 				opline->op2.u.jmp_addr = op_array->opcodes + opline->op2.u.opline_num;
 				break;
@@ -286,6 +299,9 @@ int xc_foreach_early_binding_class(zend_op_array *op_array, void (*callback)(zen
 	end = opline + op_array->last;
 	while (opline < end) {
 		switch (opline->opcode) {
+#ifdef ZEND_GOTO
+			case ZEND_GOTO:
+#endif
 			case ZEND_JMP:
 				next = begin + opline->op1.u.opline_num;
 				break;
@@ -298,6 +314,9 @@ int xc_foreach_early_binding_class(zend_op_array *op_array, void (*callback)(zen
 			case ZEND_JMPNZ:
 			case ZEND_JMPZ_EX:
 			case ZEND_JMPNZ_EX:
+#ifdef ZEND_JMP_SET
+			case ZEND_JMP_SET:
+#endif
 				next = begin + opline->op2.u.opline_num;
 				break;
 
@@ -328,13 +347,12 @@ int xc_foreach_early_binding_class(zend_op_array *op_array, void (*callback)(zen
 	return SUCCESS;
 }
 /* }}} */
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 static int xc_do_early_binding(zend_op_array *op_array, HashTable *class_table, int oplineno TSRMLS_DC) /* {{{ */
 {
 	zend_op *opline;
 
-#ifdef DEBUG
-	fprintf(stderr, "binding %d\n", oplineno);
-#endif
+	TRACE("binding %d", oplineno);
 	assert(oplineno >= 0);
 
 	/* do early binding */
@@ -353,9 +371,7 @@ static int xc_do_early_binding(zend_op_array *op_array, HashTable *class_table,
 			}
 
 			parent_name = &(opline - 1)->op2.u.constant;
-#	ifdef DEBUG
-			fprintf(stderr, "binding with parent %s\n", Z_STRVAL_P(parent_name));
-#	endif
+			TRACE("binding with parent %s", Z_STRVAL_P(parent_name));
 			if (zend_lookup_class(Z_STRVAL_P(parent_name), Z_STRLEN_P(parent_name), &pce TSRMLS_CC) == FAILURE) {
 				return FAILURE;
 			}
@@ -370,9 +386,7 @@ static int xc_do_early_binding(zend_op_array *op_array, HashTable *class_table,
 		 && (opline - 1)->opcode == ZEND_FETCH_CLASS) {
 			zend_op *fetch_class_opline = opline - 1;
 
-#	ifdef DEBUG
-			fprintf(stderr, "%s %p\n", Z_STRVAL(fetch_class_opline->op2.u.constant), Z_STRVAL(fetch_class_opline->op2.u.constant));
-#	endif
+			TRACE("%s %p", Z_STRVAL(fetch_class_opline->op2.u.constant), Z_STRVAL(fetch_class_opline->op2.u.constant));
 			OP_ZVAL_DTOR(fetch_class_opline->op2);
 			fetch_class_opline->opcode = ZEND_NOP;
 			ZEND_VM_SET_OPCODE_HANDLER(fetch_class_opline);
@@ -418,9 +432,10 @@ static int xc_do_early_binding(zend_op_array *op_array, HashTable *class_table,
 	return SUCCESS;
 }
 /* }}} */
+#endif
 
 #ifdef HAVE_XCACHE_CONSTANT
-void xc_install_constant(char *filename, zend_constant *constant, zend_uchar type, zstr key, uint len TSRMLS_DC) /* {{{ */
+void xc_install_constant(char *filename, zend_constant *constant, zend_uchar type, zstr key, uint len, ulong h TSRMLS_DC) /* {{{ */
 {
 	if (zend_u_hash_add(EG(zend_constants), type, key, len,
 				constant, sizeof(zend_constant),
@@ -440,7 +455,7 @@ void xc_install_constant(char *filename, zend_constant *constant, zend_uchar typ
 }
 /* }}} */
 #endif
-void xc_install_function(char *filename, zend_function *func, zend_uchar type, zstr key, uint len TSRMLS_DC) /* {{{ */
+void xc_install_function(char *filename, zend_function *func, zend_uchar type, zstr key, uint len, ulong h TSRMLS_DC) /* {{{ */
 {
 	zend_bool istmpkey;
 
@@ -470,7 +485,7 @@ void xc_install_function(char *filename, zend_function *func, zend_uchar type, z
 	}
 }
 /* }}} */
-ZESW(xc_cest_t *, void) xc_install_class(char *filename, xc_cest_t *cest, int oplineno, zend_uchar type, zstr key, uint len TSRMLS_DC) /* {{{ */
+ZESW(xc_cest_t *, void) xc_install_class(char *filename, xc_cest_t *cest, int oplineno, zend_uchar type, zstr key, uint len, ulong h TSRMLS_DC) /* {{{ */
 {
 	zend_bool istmpkey;
 	zend_class_entry *cep = CestToCePtr(*cest);
@@ -482,15 +497,17 @@ ZESW(xc_cest_t *, void) xc_install_class(char *filename, xc_cest_t *cest, int op
 	istmpkey = ZSTR_S(key)[0] == 0;
 #endif
 	if (istmpkey) {
-		zend_u_hash_update(CG(class_table), type, key, len,
+		zend_u_hash_quick_update(CG(class_table), type, key, len, h,
 					cest, sizeof(xc_cest_t),
 					ZESW(&stored_ce_ptr, NULL)
 					);
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 		if (oplineno != -1) {
 			xc_do_early_binding(CG(active_op_array), CG(class_table), oplineno TSRMLS_CC);
 		}
+#endif
 	}
-	else if (zend_u_hash_add(CG(class_table), type, key, len,
+	else if (zend_u_hash_quick_add(CG(class_table), type, key, len, h,
 				cest, sizeof(xc_cest_t),
 				ZESW(&stored_ce_ptr, NULL)
 				) == FAILURE) {
@@ -536,15 +553,63 @@ static int xc_auto_global_arm(zend_auto_global *auto_global TSRMLS_DC) /* {{{ */
 /* }}} */
 #endif
 
-void xc_zend_class_add_ref(zend_class_entry ZESW(*ce, **ce))
+void xc_hash_copy_if(HashTable *target, HashTable *source, copy_ctor_func_t pCopyConstructor, void *tmp, uint size, xc_if_func_t checker) /* {{{ */
 {
-#ifdef ZEND_ENGINE_2
-	(*ce)->refcount++;
+	Bucket *p;
+	void *new_entry;
+	zend_bool setTargetPointer;
+
+	setTargetPointer = !target->pInternalPointer;
+	p = source->pListHead;
+	while (p) {
+		if (checker(p->pData)) {
+			if (setTargetPointer && source->pInternalPointer == p) {
+				target->pInternalPointer = NULL;
+			}
+			if (p->nKeyLength) {
+				zend_hash_quick_update(target, p->arKey, p->nKeyLength, p->h, p->pData, size, &new_entry);
+			} else {
+				zend_hash_index_update(target, p->h, p->pData, size, &new_entry);
+			}
+			if (pCopyConstructor) {
+				pCopyConstructor(new_entry);
+			}
+		}
+		p = p->pListNext;
+	}
+	if (!target->pInternalPointer) {
+		target->pInternalPointer = target->pListHead;
+	}
+}
+/* }}} */
+#ifdef HAVE_XCACHE_CONSTANT
+static zend_bool xc_is_internal_zend_constant(zend_constant *c) /* {{{ */
+{
+	return (c->flags & CONST_PERSISTENT) ? 1 : 0;
+}
+/* }}} */
+void xc_zend_constant_ctor(zend_constant *c) /* {{{ */
+{
+	assert((c->flags & CONST_PERSISTENT));
+#ifdef IS_UNICODE
+	c->name.u = zend_ustrndup(c->name.u, c->name_len - 1);
 #else
-	(*ce->refcount)++;
+	c->name = zend_strndup(c->name, c->name_len - 1);
 #endif
 }
-
+/* }}} */
+void xc_zend_constant_dtor(zend_constant *c) /* {{{ */
+{
+	free(ZSTR_U(c->name));
+}
+/* }}} */
+void xc_copy_internal_zend_constants(HashTable *target, HashTable *source) /* {{{ */
+{
+	zend_constant tmp_const;
+	xc_hash_copy_if(target, source, (copy_ctor_func_t) xc_zend_constant_ctor, (void *) &tmp_const, sizeof(zend_constant), (xc_if_func_t) xc_is_internal_zend_constant);
+}
+/* }}} */
+#endif
 xc_sandbox_t *xc_sandbox_init(xc_sandbox_t *sandbox, char *filename TSRMLS_DC) /* {{{ */
 {
 	HashTable *h;
@@ -582,12 +647,18 @@ xc_sandbox_t *xc_sandbox_init(xc_sandbox_t *sandbox, char *filename TSRMLS_DC) /
 #ifdef HAVE_XCACHE_CONSTANT
 	h = OG(zend_constants);
 	zend_hash_init_ex(&TG(zend_constants),  20, NULL, h->pDestructor, h->persistent, h->bApplyProtection);
+	xc_copy_internal_zend_constants(&TG(zend_constants), &XG(internal_constant_table));
+	{
+		zend_constant tmp_const;
+		zend_hash_copy(&TG(zend_constants), &XG(internal_constant_table), (copy_ctor_func_t) xc_zend_constant_ctor, (void *) &tmp_const, sizeof(tmp_const));
+	}
+	TG(internal_constant_tail) = TG(zend_constants).pListTail;
 #endif
 	h = OG(function_table);
 	zend_hash_init_ex(&TG(function_table), 128, NULL, h->pDestructor, h->persistent, h->bApplyProtection);
 	{
 		zend_function tmp_func;
-		zend_hash_copy(&TG(function_table), &XG(internal_function_table), (copy_ctor_func_t) function_add_ref, (void *) &tmp_func, sizeof(tmp_func));
+		zend_hash_copy(&TG(function_table), &XG(internal_function_table), NULL, (void *) &tmp_func, sizeof(tmp_func));
 	}
 	TG(internal_function_tail) = TG(function_table).pListTail;
 
@@ -596,7 +667,7 @@ xc_sandbox_t *xc_sandbox_init(xc_sandbox_t *sandbox, char *filename TSRMLS_DC) /
 #if 0 && TODO
 	{
 		xc_cest_t tmp_cest;
-		zend_hash_copy(&TG(class_table), &XG(internal_class_table), (copy_ctor_func_t) xc_zend_class_add_ref, (void *) &tmp_cest, sizeof(tmp_cest));
+		zend_hash_copy(&TG(class_table), &XG(internal_class_table), NULL, (void *) &tmp_cest, sizeof(tmp_cest));
 	}
 #endif
 	TG(internal_class_tail) = TG(class_table).pListTail;
@@ -620,27 +691,36 @@ xc_sandbox_t *xc_sandbox_init(xc_sandbox_t *sandbox, char *filename TSRMLS_DC) /
 	EG(user_error_handler_error_reporting) &= ~E_STRICT;
 #endif
 
+#ifdef ZEND_COMPILE_IGNORE_INTERNAL_CLASSES
+	sandbox->orig_compiler_options = CG(compiler_options);
+	/* Using ZEND_COMPILE_IGNORE_INTERNAL_CLASSES for ZEND_FETCH_CLASS_RT_NS_CHECK
+	 */
+	CG(compiler_options) |= ZEND_COMPILE_IGNORE_INTERNAL_CLASSES | ZEND_COMPILE_NO_CONSTANT_SUBSTITUTION | ZEND_COMPILE_DELAYED_BINDING;
+#endif
+
 	return sandbox;
 }
 /* }}} */
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 static void xc_early_binding_cb(zend_op *opline, int oplineno, void *data TSRMLS_DC) /* {{{ */
 {
 	xc_sandbox_t *sandbox = (xc_sandbox_t *) data;
 	xc_do_early_binding(CG(active_op_array), OG(class_table), oplineno TSRMLS_CC);
 }
 /* }}} */
+#endif
 static void xc_sandbox_install(xc_sandbox_t *sandbox, xc_install_action_t install TSRMLS_DC) /* {{{ */
 {
-	int i;
+	zend_uint i;
 	Bucket *b;
 
 #ifdef HAVE_XCACHE_CONSTANT
-	b = TG(zend_constants).pListHead;
+	b = TG(internal_constant_tail) ? TG(internal_constant_tail)->pListNext : TG(zend_constants).pListHead;
 	/* install constants */
 	while (b != NULL) {
 		zend_constant *c = (zend_constant*) b->pData;
 		xc_install_constant(sandbox->filename, c,
-				BUCKET_KEY_TYPE(b), ZSTR(BUCKET_KEY_S(b)), b->nKeyLength TSRMLS_CC);
+				BUCKET_KEY_TYPE(b), ZSTR(BUCKET_KEY_S(b)), b->nKeyLength, b->h TSRMLS_CC);
 		b = b->pListNext;
 	}
 #endif
@@ -650,7 +730,7 @@ static void xc_sandbox_install(xc_sandbox_t *sandbox, xc_install_action_t instal
 	while (b != NULL) {
 		zend_function *func = (zend_function*) b->pData;
 		xc_install_function(sandbox->filename, func,
-				BUCKET_KEY_TYPE(b), ZSTR(BUCKET_KEY_S(b)), b->nKeyLength TSRMLS_CC);
+				BUCKET_KEY_TYPE(b), ZSTR(BUCKET_KEY_S(b)), b->nKeyLength, b->h TSRMLS_CC);
 		b = b->pListNext;
 	}
 
@@ -658,7 +738,7 @@ static void xc_sandbox_install(xc_sandbox_t *sandbox, xc_install_action_t instal
 	/* install class */
 	while (b != NULL) {
 		xc_install_class(sandbox->filename, (xc_cest_t*) b->pData, -1,
-				BUCKET_KEY_TYPE(b), ZSTR(BUCKET_KEY_S(b)), b->nKeyLength TSRMLS_CC);
+				BUCKET_KEY_TYPE(b), ZSTR(BUCKET_KEY_S(b)), b->nKeyLength, b->h TSRMLS_CC);
 		b = b->pListNext;
 	}
 
@@ -674,9 +754,13 @@ static void xc_sandbox_install(xc_sandbox_t *sandbox, xc_install_action_t instal
 #endif
 
 	if (install != XC_InstallNoBinding) {
+#ifdef ZEND_COMPILE_DELAYED_BINDING
+		zend_do_delayed_early_binding(CG(active_op_array) TSRMLS_CC);
+#else
 		xc_undo_pass_two(CG(active_op_array) TSRMLS_CC);
 		xc_foreach_early_binding_class(CG(active_op_array), xc_early_binding_cb, (void *) sandbox TSRMLS_CC);
 		xc_redo_pass_two(CG(active_op_array) TSRMLS_CC);
+#endif
 	}
 
 	i = 1;
@@ -730,8 +814,28 @@ void xc_sandbox_free(xc_sandbox_t *sandbox, xc_install_action_t install TSRMLS_D
 	EG(user_error_handler_error_reporting) = sandbox->orig_user_error_handler_error_reporting;
 #endif
 
+#ifdef ZEND_COMPILE_IGNORE_INTERNAL_CLASSES
+	CG(compiler_options) = sandbox->orig_compiler_options;
+#endif
+
 	if (sandbox->alloc) {
 		efree(sandbox);
 	}
 }
 /* }}} */
+int xc_vtrace(const char *fmt, va_list args) /* {{{ */
+{
+	return vfprintf(stderr, fmt, args);
+}
+/* }}} */
+int xc_trace(const char *fmt, ...) /* {{{ */
+{
+	va_list args;
+	int ret;
+
+	va_start(args, fmt);
+	ret = xc_vtrace(fmt, args);
+	va_end(args);
+	return ret;
+}
+/* }}} */
