@@ -1,6 +1,6 @@
 
 #if 0
-#define DEBUG
+#define XCACHE_DEBUG
 #endif
 
 #if 0
@@ -22,6 +22,9 @@
 #include "SAPI.h"
 
 #include "xcache.h"
+#ifdef ZEND_ENGINE_2_1
+#include "ext/date/php_date.h"
+#endif
 #include "optimizer.h"
 #include "coverager.h"
 #include "disassembler.h"
@@ -29,20 +32,9 @@
 #include "stack.h"
 #include "xcache_globals.h"
 #include "processor.h"
-#include "utils.h"
 #include "const_string.h"
 #include "opcode_spec.h"
-
-#ifdef DEBUG
-#	undef NDEBUG
-#	undef inline
-#	define inline
-#else
-#	ifndef NDEBUG
-#		define NDEBUG
-#	endif
-#endif
-#include <assert.h>
+#include "utils.h"
 
 #define VAR_ENTRY_EXPIRED(pentry) ((pentry)->ttl && XG(request_time) > pentry->ctime + (pentry)->ttl)
 #define CHECK(x, e) do { if ((x) == NULL) { zend_error(E_ERROR, "XCache: " e); goto err; } } while (0)
@@ -97,6 +89,11 @@ static xc_cache_t **xc_php_caches = NULL;
 static xc_cache_t **xc_var_caches = NULL;
 
 static zend_bool xc_initized = 0;
+static time_t xc_init_time = 0;
+static long unsigned xc_init_instance_id = 0;
+#ifdef ZTS
+static long unsigned xc_init_instance_subid = 0;
+#endif
 static zend_compile_file_t *origin_compile_file = NULL;
 static zend_compile_file_t *old_compile_file = NULL;
 static zend_llist_element  *xc_llist_zend_extension = NULL;
@@ -228,7 +225,7 @@ static xc_entry_t *xc_entry_find_dmz(xc_entry_t *xce TSRMLS_DC) /* {{{ */
 	xc_entry_t *p;
 	for (p = xce->cache->entries[xce->hvalue]; p; p = p->next) {
 		if (xc_entry_equal_dmz(xce, p)) {
-			if (p->type == XC_TYPE_VAR || /* PHP */ p->data.php->mtime == xce->data.php->mtime) {
+			if (p->type == XC_TYPE_VAR || /* PHP */ (p->data.php->mtime == xce->data.php->mtime && p->data.php->sourcesize == xce->data.php->sourcesize)) {
 				p->hits ++;
 				p->atime = XG(request_time);
 				return p;
@@ -244,9 +241,7 @@ static xc_entry_t *xc_entry_find_dmz(xc_entry_t *xce TSRMLS_DC) /* {{{ */
 /* }}} */
 static void xc_entry_hold_php_dmz(xc_entry_t *xce TSRMLS_DC) /* {{{ */
 {
-#ifdef DEBUG
-	fprintf(stderr, "hold %s\n", xce->name.str.val);
-#endif
+	TRACE("hold %s", xce->name.str.val);
 	xce->refcount ++;
 	xc_stack_push(&XG(php_holds)[xce->cache->cacheid], (void *)xce);
 }
@@ -259,6 +254,55 @@ static void xc_entry_hold_var_dmz(xc_entry_t *xce TSRMLS_DC) /* {{{ */
 }
 /* }}} */
 #endif
+static inline zend_uint advance_wrapped(zend_uint val, zend_uint count) /* {{{ */
+{
+	if (val + 1 >= count) {
+		return 0;
+	}
+	return val + 1;
+}
+/* }}} */
+static void xc_counters_inc(time_t *curtime, zend_uint *curslot, time_t period, zend_ulong *counters, zend_uint count TSRMLS_DC) /* {{{ */
+{
+	time_t n = XG(request_time) / period;
+	if (*curtime != n) {
+		zend_uint target_slot = n % count;
+		if (n - *curtime > period) {
+			memset(counters, 0, sizeof(counters[0]) * count);
+		}
+		else {
+			zend_uint slot;
+			for (slot = advance_wrapped(*curslot, count);
+					slot != target_slot;
+					slot = advance_wrapped(slot, count)) {
+				counters[slot] = 0;
+			}
+			counters[target_slot] = 0;
+		}
+		*curtime = n;
+		*curslot = target_slot;
+	}
+	counters[*curslot] ++;
+}
+/* }}} */
+static void xc_cache_hit_dmz(xc_cache_t *cache TSRMLS_DC) /* {{{ */
+{
+	cache->hits ++;
+
+	xc_counters_inc(&cache->hits_by_hour_cur_time
+			, &cache->hits_by_hour_cur_slot, 60 * 60
+			, cache->hits_by_hour
+			, sizeof(cache->hits_by_hour) / sizeof(cache->hits_by_hour[0])
+			TSRMLS_CC);
+
+	xc_counters_inc(&cache->hits_by_second_cur_time
+			, &cache->hits_by_second_cur_slot
+			, 1
+			, cache->hits_by_second
+			, sizeof(cache->hits_by_second) / sizeof(cache->hits_by_second[0])
+			TSRMLS_CC);
+}
+/* }}} */
 
 /* helper function that loop through each entry */
 #define XC_ENTRY_APPLY_FUNC(name) int name(xc_entry_t *entry TSRMLS_DC)
@@ -291,9 +335,7 @@ static void xc_entry_apply_dmz(xc_cache_t *cache, cache_apply_dmz_func_t apply_f
  */
 static XC_ENTRY_APPLY_FUNC(xc_gc_expires_php_entry_dmz) /* {{{ */
 {
-#ifdef DEBUG
-	fprintf(stderr, "ttl %lu, %lu %lu\n", (zend_ulong) XG(request_time), (zend_ulong) entry->atime, xc_php_ttl);
-#endif
+	TRACE("ttl %lu, %lu %lu", (zend_ulong) XG(request_time), (zend_ulong) entry->atime, xc_php_ttl);
 	if (XG(request_time) > entry->atime + xc_php_ttl) {
 		return 1;
 	}
@@ -310,9 +352,7 @@ static XC_ENTRY_APPLY_FUNC(xc_gc_expires_var_entry_dmz) /* {{{ */
 /* }}} */
 static void xc_gc_expires_one(xc_cache_t *cache, zend_ulong gc_interval, cache_apply_dmz_func_t apply_func TSRMLS_DC) /* {{{ */
 {
-#ifdef DEBUG
-	fprintf(stderr, "interval %lu, %lu %lu\n", (zend_ulong) XG(request_time), (zend_ulong) cache->last_gc_expires, gc_interval);
-#endif
+	TRACE("interval %lu, %lu %lu", (zend_ulong) XG(request_time), (zend_ulong) cache->last_gc_expires, gc_interval);
 	if (XG(request_time) - cache->last_gc_expires >= gc_interval) {
 		ENTER_LOCK(cache) {
 			if (XG(request_time) - cache->last_gc_expires >= gc_interval) {
@@ -327,7 +367,7 @@ static void xc_gc_expires_php(TSRMLS_D) /* {{{ */
 {
 	int i, c;
 
-	if (!xc_php_ttl || !xc_php_gc_interval) {
+	if (!xc_php_ttl || !xc_php_gc_interval || !xc_php_caches) {
 		return;
 	}
 
@@ -340,7 +380,7 @@ static void xc_gc_expires_var(TSRMLS_D) /* {{{ */
 {
 	int i, c;
 
-	if (!xc_var_gc_interval) {
+	if (!xc_var_gc_interval || !xc_var_caches) {
 		return;
 	}
 
@@ -388,12 +428,16 @@ static void xc_gc_deletes(TSRMLS_D) /* {{{ */
 {
 	int i, c;
 
-	for (i = 0, c = xc_php_hcache.size; i < c; i ++) {
-		xc_gc_deletes_one(xc_php_caches[i] TSRMLS_CC);
+	if (xc_php_caches) {
+		for (i = 0, c = xc_php_hcache.size; i < c; i ++) {
+			xc_gc_deletes_one(xc_php_caches[i] TSRMLS_CC);
+		}
 	}
 
-	for (i = 0, c = xc_var_hcache.size; i < c; i ++) {
-		xc_gc_deletes_one(xc_var_caches[i] TSRMLS_CC);
+	if (xc_var_caches) {
+		for (i = 0, c = xc_var_hcache.size; i < c; i ++) {
+			xc_gc_deletes_one(xc_var_caches[i] TSRMLS_CC);
+		}
 	}
 }
 /* }}} */
@@ -401,14 +445,21 @@ static void xc_gc_deletes(TSRMLS_D) /* {{{ */
 /* helper functions for user functions */
 static void xc_fillinfo_dmz(int cachetype, xc_cache_t *cache, zval *return_value TSRMLS_DC) /* {{{ */
 {
-	zval *blocks;
+	zval *blocks, *hits;
+	int i;
 	const xc_block_t *b;
 #ifndef NDEBUG
 	xc_memsize_t avail = 0;
 #endif
 	xc_mem_t *mem = cache->mem;
 	const xc_mem_handlers_t *handlers = mem->handlers;
-	zend_ulong interval = (cachetype == XC_TYPE_PHP) ? xc_php_gc_interval : xc_var_gc_interval;
+	zend_ulong interval;
+	if (cachetype == XC_TYPE_PHP) {
+		interval = xc_php_ttl ? xc_php_gc_interval : 0;
+	}
+	else {
+		interval = xc_var_gc_interval;
+	}
 
 	add_assoc_long_ex(return_value, ZEND_STRS("slots"),     cache->hentry->size);
 	add_assoc_long_ex(return_value, ZEND_STRS("compiling"), cache->compiling);
@@ -421,11 +472,25 @@ static void xc_fillinfo_dmz(int cachetype, xc_cache_t *cache, zval *return_value
 	add_assoc_long_ex(return_value, ZEND_STRS("cached"),    cache->entries_count);
 	add_assoc_long_ex(return_value, ZEND_STRS("deleted"),   cache->deletes_count);
 	if (interval) {
-		add_assoc_long_ex(return_value, ZEND_STRS("gc"),    (cache->last_gc_expires + interval) - XG(request_time));
+		time_t gc = (cache->last_gc_expires + interval) - XG(request_time);
+		add_assoc_long_ex(return_value, ZEND_STRS("gc"),    gc > 0 ? gc : 0);
 	}
 	else {
 		add_assoc_null_ex(return_value, ZEND_STRS("gc"));
 	}
+	MAKE_STD_ZVAL(hits);
+	array_init(hits);
+	for (i = 0; i < sizeof(cache->hits_by_hour) / sizeof(cache->hits_by_hour[0]); i ++) {
+		add_next_index_long(hits, (long) cache->hits_by_hour[i]);
+	}
+	add_assoc_zval_ex(return_value, ZEND_STRS("hits_by_hour"), hits);
+
+	MAKE_STD_ZVAL(hits);
+	array_init(hits);
+	for (i = 0; i < sizeof(cache->hits_by_second) / sizeof(cache->hits_by_second[0]); i ++) {
+		add_next_index_long(hits, (long) cache->hits_by_second[i]);
+	}
+	add_assoc_zval_ex(return_value, ZEND_STRS("hits_by_second"), hits);
 
 	MAKE_STD_ZVAL(blocks);
 	array_init(blocks);
@@ -448,7 +513,9 @@ static void xc_fillinfo_dmz(int cachetype, xc_cache_t *cache, zval *return_value
 #endif
 	}
 	add_assoc_zval_ex(return_value, ZEND_STRS("free_blocks"), blocks);
+#ifndef NDEBUG
 	assert(avail == handlers->avail(mem));
+#endif
 }
 /* }}} */
 static void xc_fillentry_dmz(xc_entry_t *entry, int del, zval *list TSRMLS_DC) /* {{{ */
@@ -465,6 +532,7 @@ static void xc_fillentry_dmz(xc_entry_t *entry, int del, zval *list TSRMLS_DC) /
 	add_assoc_long_ex(ei, ZEND_STRS("hits"),     entry->hits);
 	add_assoc_long_ex(ei, ZEND_STRS("ctime"),    entry->ctime);
 	add_assoc_long_ex(ei, ZEND_STRS("atime"),    entry->atime);
+	add_assoc_long_ex(ei, ZEND_STRS("hvalue"),   entry->hvalue);
 	if (del) {
 		add_assoc_long_ex(ei, ZEND_STRS("dtime"), entry->dtime);
 	}
@@ -549,8 +617,9 @@ static zend_op_array *xc_entry_install(xc_entry_t *xce, zend_file_handle *h TSRM
 	xc_entry_data_php_t *p = xce->data.php;
 	zend_op_array *old_active_op_array = CG(active_op_array);
 #ifndef ZEND_ENGINE_2
+	ALLOCA_FLAG(use_heap)
 	/* new ptr which is stored inside CG(class_table) */
-	xc_cest_t **new_cest_ptrs = (xc_cest_t **)do_alloca(sizeof(xc_cest_t*) * p->classinfo_cnt);
+	xc_cest_t **new_cest_ptrs = (xc_cest_t **)my_do_alloca(sizeof(xc_cest_t*) * p->classinfo_cnt, use_heap);
 #endif
 
 	CG(active_op_array) = p->op_array;
@@ -560,7 +629,7 @@ static zend_op_array *xc_entry_install(xc_entry_t *xce, zend_file_handle *h TSRM
 	for (i = 0; i < p->constinfo_cnt; i ++) {
 		xc_constinfo_t *ci = &p->constinfos[i];
 		xc_install_constant(xce->name.str.val, &ci->constant,
-				UNISW(0, ci->type), ci->key, ci->key_size TSRMLS_CC);
+				UNISW(0, ci->type), ci->key, ci->key_size, ci->h TSRMLS_CC);
 	}
 #endif
 
@@ -568,7 +637,7 @@ static zend_op_array *xc_entry_install(xc_entry_t *xce, zend_file_handle *h TSRM
 	for (i = 0; i < p->funcinfo_cnt; i ++) {
 		xc_funcinfo_t  *fi = &p->funcinfos[i];
 		xc_install_function(xce->name.str.val, &fi->func,
-				UNISW(0, fi->type), fi->key, fi->key_size TSRMLS_CC);
+				UNISW(0, fi->type), fi->key, fi->key_size, fi->h TSRMLS_CC);
 	}
 
 	/* install class */
@@ -584,8 +653,13 @@ static zend_op_array *xc_entry_install(xc_entry_t *xce, zend_file_handle *h TSRM
 		}
 		new_cest_ptrs[i] =
 #endif
+#ifdef ZEND_COMPILE_DELAYED_BINDING
+		xc_install_class(xce->name.str.val, &ci->cest, -1,
+				UNISW(0, ci->type), ci->key, ci->key_size, ci->h TSRMLS_CC);
+#else
 		xc_install_class(xce->name.str.val, &ci->cest, ci->oplineno,
-				UNISW(0, ci->type), ci->key, ci->key_size TSRMLS_CC);
+				UNISW(0, ci->type), ci->key, ci->key_size, ci->h TSRMLS_CC);
+#endif
 	}
 
 #ifdef ZEND_ENGINE_2_1
@@ -594,7 +668,7 @@ static zend_op_array *xc_entry_install(xc_entry_t *xce, zend_file_handle *h TSRM
 		xc_autoglobal_t *aginfo = &p->autoglobals[i];
 		/*
 		zend_auto_global *auto_global;
-		if (zend_u_hash_find(CG(auto_globals), aginfo->type, aginfo->key, aginfo->key_len+1, (void **) &auto_global)==SUCCESS) {
+		if (zend_u_hash_quick_find(CG(auto_globals), aginfo->type, aginfo->key, aginfo->key_len+1, aginfo->h, (void **) &auto_global)==SUCCESS) {
 			if (auto_global->armed) {
 				auto_global->armed = auto_global->auto_global_callback(auto_global->name, auto_global->name_len TSRMLS_CC);
 			}
@@ -611,7 +685,7 @@ static zend_op_array *xc_entry_install(xc_entry_t *xce, zend_file_handle *h TSRM
 	}
 
 #ifndef ZEND_ENGINE_2
-	free_alloca(new_cest_ptrs);
+	my_free_alloca(new_cest_ptrs, use_heap);
 #endif
 	CG(active_op_array) = old_active_op_array;
 	return p->op_array;
@@ -627,17 +701,13 @@ static inline void xc_entry_unholds_real(xc_stack_t *holds, xc_cache_t **caches,
 
 	for (i = 0; i < cachecount; i ++) {
 		s = &holds[i];
-#ifdef DEBUG
-		fprintf(stderr, "holded %d\n", xc_stack_size(s));
-#endif
-		if (xc_stack_size(s)) {
+		TRACE("holded %d", xc_stack_count(s));
+		if (xc_stack_count(s)) {
 			cache = ((xc_entry_t *)xc_stack_top(s))->cache;
 			ENTER_LOCK(cache) {
-				while (xc_stack_size(s)) {
+				while (xc_stack_count(s)) {
 					xce = (xc_entry_t*) xc_stack_pop(s);
-#ifdef DEBUG
-					fprintf(stderr, "unhold %s\n", xce->name.str.val);
-#endif
+					TRACE("unhold %s", xce->name.str.val);
 					xce->refcount --;
 					assert(xce->refcount >= 0);
 				}
@@ -648,8 +718,13 @@ static inline void xc_entry_unholds_real(xc_stack_t *holds, xc_cache_t **caches,
 /* }}} */
 static void xc_entry_unholds(TSRMLS_D) /* {{{ */
 {
-	xc_entry_unholds_real(XG(php_holds), xc_php_caches, xc_php_hcache.size TSRMLS_CC);
-	xc_entry_unholds_real(XG(var_holds), xc_var_caches, xc_var_hcache.size TSRMLS_CC);
+	if (xc_php_caches) {
+		xc_entry_unholds_real(XG(php_holds), xc_php_caches, xc_php_hcache.size TSRMLS_CC);
+	}
+
+	if (xc_var_caches) {
+		xc_entry_unholds_real(XG(var_holds), xc_var_caches, xc_var_hcache.size TSRMLS_CC);
+	}
 }
 /* }}} */
 static int xc_stat(const char *filename, const char *include_path, struct stat *pbuf TSRMLS_DC) /* {{{ */
@@ -659,8 +734,9 @@ static int xc_stat(const char *filename, const char *include_path, struct stat *
 	char *tokbuf;
 	int size = strlen(include_path) + 1;
 	char tokens[] = { DEFAULT_DIR_SEPARATOR, '\0' };
+	ALLOCA_FLAG(use_heap)
 
-	paths = (char *)do_alloca(size);
+	paths = (char *)my_do_alloca(size, use_heap);
 	memcpy(paths, include_path, size);
 
 	for (path = php_strtok_r(paths, tokens, &tokbuf); path; path = php_strtok_r(NULL, tokens, &tokbuf)) {
@@ -668,19 +744,19 @@ static int xc_stat(const char *filename, const char *include_path, struct stat *
 			continue;
 		}
 		if (VCWD_STAT(filepath, pbuf) == 0) {
-			free_alloca(paths);
-			return 0;
+			my_free_alloca(paths, use_heap);
+			return SUCCESS;
 		}
 	}
 
-	free_alloca(paths);
+	my_free_alloca(paths, use_heap);
 
-	return 1;
+	return FAILURE;
 }
 /* }}} */
 
 #define HASH(i) (i)
-#define HASH_USTR_L(t, s, l) HASH(zend_u_inline_hash_func(t, s, (l + 1) * sizeof(UChar)))
+#define HASH_ZSTR_L(t, s, l) HASH(zend_u_inline_hash_func(t, s, (l + 1) * sizeof(UChar)))
 #define HASH_STR_L(s, l) HASH(zend_inline_hash_func(s, l + 1))
 #define HASH_STR(s) HASH_STR_L(s, strlen(s) + 1)
 #define HASH_NUM(n) HASH(n)
@@ -696,7 +772,7 @@ static inline xc_hash_value_t xc_hash_fold(xc_hash_value_t hvalue, const xc_hash
 /* }}} */
 static inline xc_hash_value_t xc_entry_hash_name(xc_entry_t *xce TSRMLS_DC) /* {{{ */
 {
-	return UNISW(NOTHING, UG(unicode) ? HASH_USTR_L(xce->name_type, xce->name.uni.val, xce->name.uni.len) :)
+	return UNISW(NOTHING, UG(unicode) ? HASH_ZSTR_L(xce->name_type, xce->name.uni.val, xce->name.uni.len) :)
 		HASH_STR_L(xce->name.str.val, xce->name.str.len);
 }
 /* }}} */
@@ -721,7 +797,11 @@ static int xc_entry_init_key_php(xc_entry_t *xce, char *filename, char *opened_p
 	time_t delta;
 
 	if (!filename || !SG(request_info).path_translated) {
-		return 0;
+		return FAILURE;
+	}
+
+	if (strstr(filename, "://") != NULL) {
+		return FAILURE;
 	}
 
 	php = xce->data.php;
@@ -739,7 +819,7 @@ static int xc_entry_init_key_php(xc_entry_t *xce, char *filename, char *opened_p
 		pbuf = &buf;
 		if (IS_ABSOLUTE_PATH(filename, strlen(filename))) {
 			if (VCWD_STAT(filename, pbuf) != 0) {
-				return 0;
+				return FAILURE;
 			}
 			goto stat_done;
 		}
@@ -755,15 +835,15 @@ static int xc_entry_init_key_php(xc_entry_t *xce, char *filename, char *opened_p
 			}
 
 			if (VCWD_STAT(filename, pbuf) != 0) {
-				return 0;
+				return FAILURE;
 			}
 			goto stat_done;
 		}
 not_relative_path:
 
 		/* use include_path */
-		if (xc_stat(filename, PG(include_path), pbuf TSRMLS_CC) != 0) {   
-			return 0;
+		if (xc_stat(filename, PG(include_path), pbuf TSRMLS_CC) != SUCCESS) {
+			return FAILURE;
 		}
 
 		/* fall */
@@ -771,7 +851,7 @@ not_relative_path:
 stat_done:
 		delta = XG(request_time) - pbuf->st_mtime;
 		if (abs(delta) < 2 && !xc_test) {
-			return 0;
+			return FAILURE;
 		}
 
 		php->mtime        = pbuf->st_mtime;
@@ -797,7 +877,7 @@ stat_done:
 		/* hash on filename, let's expand it to real path */
 		filename = expand_filepath(filename, opened_path_buffer TSRMLS_CC);
 		if (filename == NULL) {
-			return 0;
+			return FAILURE;
 		}
 	}
 
@@ -811,9 +891,10 @@ stat_done:
 	xce->hvalue = xc_hash_fold(hv, &xc_php_hentry);
 
 	xce->type = XC_TYPE_PHP;
-	return 1;
+	return SUCCESS;
 }
 /* }}} */
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 static void xc_cache_early_binding_class_cb(zend_op *opline, int oplineno, void *data TSRMLS_DC) /* {{{ */
 {
 	char *class_name;
@@ -826,9 +907,7 @@ static void xc_cache_early_binding_class_cb(zend_op *opline, int oplineno, void
 	if (zend_hash_find(CG(class_table), class_name, class_len, (void **) &cest) == FAILURE) {
 		assert(0);
 	}
-#ifdef DEBUG
-	fprintf(stderr, "got ZEND_DECLARE_INHERITED_CLASS: %s\n", class_name + 1);
-#endif
+	TRACE("got ZEND_DECLARE_INHERITED_CLASS: %s", class_name + 1);
 	/* let's see which class */
 	for (i = 0; i < php->classinfo_cnt; i ++) {
 		if (memcmp(ZSTR_S(php->classinfos[i].key), class_name, class_len) == 0) {
@@ -843,6 +922,7 @@ static void xc_cache_early_binding_class_cb(zend_op *opline, int oplineno, void
 	}
 }
 /* }}} */
+#endif
 static zend_op_array *xc_check_initial_compile_file(zend_file_handle *h, int type TSRMLS_DC) /* {{{ */
 {
 	XG(initial_compile_file_called) = 1;
@@ -861,7 +941,6 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 	char *filename;
 	char opened_path_buffer[MAXPATHLEN];
 	int old_constinfo_cnt, old_funcinfo_cnt, old_classinfo_cnt;
-	int i;
 
 	if (!xc_initized) {
 		assert(0);
@@ -886,7 +965,7 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 
 	filename = h->opened_path ? h->opened_path : h->filename;
 	xce.data.php = &php;
-	if (!xc_entry_init_key_php(&xce, filename, opened_path_buffer TSRMLS_CC)) {
+	if (xc_entry_init_key_php(&xce, filename, opened_path_buffer TSRMLS_CC) != SUCCESS) {
 		return old_compile_file(h, type TSRMLS_CC);
 	}
 	cache = xce.cache;
@@ -912,11 +991,9 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 		stored_xce = xc_entry_find_dmz(&xce TSRMLS_CC);
 		/* found */
 		if (stored_xce) {
-#ifdef DEBUG
-			fprintf(stderr, "found %s, catch it\n", stored_xce->name.str.val);
-#endif
+			TRACE("found %s, catch it", stored_xce->name.str.val);
 			xc_entry_hold_php_dmz(stored_xce TSRMLS_CC);
-			cache->hits ++;
+			xc_cache_hit_dmz(cache TSRMLS_CC);
 			break;
 		}
 
@@ -941,9 +1018,7 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 	/* }}} */
 
 	/* {{{ compile */
-#ifdef DEBUG
-	fprintf(stderr, "compiling %s\n", filename);
-#endif
+	TRACE("compiling %s", filename);
 
 	/* make compile inside sandbox */
 	xc_sandbox_init(&sandbox, filename TSRMLS_CC);
@@ -1068,6 +1143,7 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 			ZSTR_U(data->key)      = BUCKET_KEY_U(b);         \
 		}                                                     \
 		data->key_size   = b->nKeyLength;                     \
+		data->h          = b->h;                              \
 	}                                                         \
 } while(0)
 
@@ -1100,21 +1176,28 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 					ZSTR_U(data->key)     = BUCKET_KEY_U(b);
 				}
 				data->key_len = b->nKeyLength - 1;
+				data->h       = b->h;
 			}
 		}
 #endif
 	}
 	/* }}} */
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 	/* {{{ find inherited classes that should be early-binding */
 	php.have_early_binding = 0;
-	for (i = 0; i < php.classinfo_cnt; i ++) {
-		php.classinfos[i].oplineno = -1;
+	{
+		int i;
+		for (i = 0; i < php.classinfo_cnt; i ++) {
+			php.classinfos[i].oplineno = -1;
+		}
 	}
 
 	xc_undo_pass_two(php.op_array TSRMLS_CC);
 	xc_foreach_early_binding_class(php.op_array, xc_cache_early_binding_class_cb, (void *) &php TSRMLS_CC);
 	xc_redo_pass_two(php.op_array TSRMLS_CC);
 	/* }}} */
+#endif
+
 #ifdef SHOW_DPRINT
 	xc_dprint(&xce, 0 TSRMLS_CC);
 #endif
@@ -1122,9 +1205,7 @@ static zend_op_array *xc_compile_file(zend_file_handle *h, int type TSRMLS_DC) /
 		stored_xce = xc_entry_store_dmz(&xce TSRMLS_CC);
 	} LEAVE_LOCK_EX(cache);
 	/* }}} */
-#ifdef DEBUG
-	fprintf(stderr, "stored\n");
-#endif
+	TRACE("%s", "stored");
 
 #define X_FREE(var) \
 	if (xce.data.php->var) { \
@@ -1183,9 +1264,7 @@ restore:
 	CG(in_compilation)    = 1;
 	CG(compiled_filename) = stored_xce->name.str.val;
 	CG(zend_lineno)       = 0;
-#ifdef DEBUG
-	fprintf(stderr, "restoring %s\n", stored_xce->name.str.val);
-#endif
+	TRACE("restoring %s", stored_xce->name.str.val);
 	xc_processor_restore_xc_entry_t(&xce, stored_xce, xc_readonly_protection TSRMLS_CC);
 #ifdef SHOW_DPRINT
 	xc_dprint(&xce, 0 TSRMLS_CC);
@@ -1202,9 +1281,6 @@ restore:
 	if (xce.data.php->var) { \
 		efree(xce.data.php->var); \
 	}
-#ifdef ZEND_ENGINE_2_1
-	X_FREE(autoglobals)
-#endif
 	X_FREE(classinfos)
 	X_FREE(funcinfos)
 #ifdef HAVE_XCACHE_CONSTANT
@@ -1218,9 +1294,7 @@ restore:
 	}
 	CG(in_compilation)    = 0;
 	CG(compiled_filename) = NULL;
-#ifdef DEBUG
-	fprintf(stderr, "restored  %s\n", stored_xce->name.str.val);
-#endif
+	TRACE("restored  %s", stored_xce->name.str.val);
 	return op_array;
 }
 /* }}} */
@@ -1230,19 +1304,22 @@ int xc_is_rw(const void *p) /* {{{ */
 {
 	xc_shm_t *shm;
 	int i;
-	if (!xc_initized) {
-		return 0;
-	}
-	for (i = 0; i < xc_php_hcache.size; i ++) {
-		shm = xc_php_caches[i]->shm;
-		if (shm->handlers->is_readwrite(shm, p)) {
-			return 1;
+
+	if (xc_php_caches) {
+		for (i = 0; i < xc_php_hcache.size; i ++) {
+			shm = xc_php_caches[i]->shm;
+			if (shm->handlers->is_readwrite(shm, p)) {
+				return 1;
+			}
 		}
 	}
-	for (i = 0; i < xc_var_hcache.size; i ++) {
-		shm = xc_var_caches[i]->shm;
-		if (shm->handlers->is_readwrite(shm, p)) {
-			return 1;
+
+	if (xc_var_caches) {
+		for (i = 0; i < xc_var_hcache.size; i ++) {
+			shm = xc_var_caches[i]->shm;
+			if (shm->handlers->is_readwrite(shm, p)) {
+				return 1;
+			}
 		}
 	}
 	return 0;
@@ -1252,19 +1329,22 @@ int xc_is_ro(const void *p) /* {{{ */
 {
 	xc_shm_t *shm;
 	int i;
-	if (!xc_initized) {
-		return 0;
-	}
-	for (i = 0; i < xc_php_hcache.size; i ++) {
-		shm = xc_php_caches[i]->shm;
-		if (shm->handlers->is_readonly(shm, p)) {
-			return 1;
+
+	if (xc_php_caches) {
+		for (i = 0; i < xc_php_hcache.size; i ++) {
+			shm = xc_php_caches[i]->shm;
+			if (shm->handlers->is_readonly(shm, p)) {
+				return 1;
+			}
 		}
 	}
-	for (i = 0; i < xc_var_hcache.size; i ++) {
-		shm = xc_var_caches[i]->shm;
-		if (shm->handlers->is_readonly(shm, p)) {
-			return 1;
+
+	if (xc_var_caches) {
+		for (i = 0; i < xc_var_hcache.size; i ++) {
+			shm = xc_var_caches[i]->shm;
+			if (shm->handlers->is_readonly(shm, p)) {
+				return 1;
+			}
 		}
 	}
 	return 0;
@@ -1276,6 +1356,40 @@ int xc_is_shm(const void *p) /* {{{ */
 }
 /* }}} */
 
+#ifdef ZEND_ENGINE_2
+/* {{{ xc_gc_op_array_t */
+typedef struct {
+	zend_uint num_args;
+	zend_arg_info *arg_info;
+} xc_gc_op_array_t;
+/* }}} */
+void xc_gc_add_op_array(zend_op_array *op_array TSRMLS_DC) /* {{{ */
+{
+	xc_gc_op_array_t gc_op_array;
+	gc_op_array.num_args = op_array->num_args;
+	gc_op_array.arg_info = op_array->arg_info;
+#ifdef ZEND_ENGINE_2
+	zend_hash_next_index_insert(&XG(gc_op_arrays), (void *) &gc_op_array, sizeof(gc_op_array), NULL);
+#endif
+}
+/* }}} */
+static void xc_gc_op_array(void *pDest) /* {{{ */
+{
+	xc_gc_op_array_t *op_array = (xc_gc_op_array_t *) pDest;
+	zend_uint i;
+	if (op_array->arg_info) {
+		for (i = 0; i < op_array->num_args; i++) {
+			efree((char *) ZSTR_V(op_array->arg_info[i].name));
+			if (ZSTR_V(op_array->arg_info[i].class_name)) {
+				efree((char *) ZSTR_V(op_array->arg_info[i].class_name));
+			}
+		}
+		efree(op_array->arg_info);
+	}
+}
+/* }}} */
+#endif
+
 /* module helper function */
 static int xc_init_constant(int module_number TSRMLS_DC) /* {{{ */
 {
@@ -1414,13 +1528,11 @@ static void xc_destroy() /* {{{ */
 		shm = xc_cache_destroy(xc_php_caches, &xc_php_hcache);
 		xc_php_caches = NULL;
 	}
-	xc_php_hcache.size = 0;
 
 	if (xc_var_caches) {
 		shm = xc_cache_destroy(xc_var_caches, &xc_var_hcache);
 		xc_var_caches = NULL;
 	}
-	xc_var_hcache.size = 0;
 
 	if (shm) {
 		xc_shm_destroy(shm);
@@ -1462,12 +1574,11 @@ static int xc_init(int module_number TSRMLS_DC) /* {{{ */
 	return SUCCESS;
 
 err:
+	xc_destroy();
 	if (xc_php_caches || xc_var_caches) {
-		xc_destroy();
 		/* shm destroied in xc_destroy() */
 	}
 	else if (shm) {
-		xc_destroy();
 		xc_shm_destroy(shm);
 	}
 	return 0;
@@ -1481,31 +1592,44 @@ static void xc_request_init(TSRMLS_D) /* {{{ */
 		zend_function tmp_func;
 		xc_cest_t tmp_cest;
 
+#ifdef HAVE_XCACHE_CONSTANT
+		zend_hash_destroy(&XG(internal_constant_table));
+#endif
 		zend_hash_destroy(&XG(internal_function_table));
 		zend_hash_destroy(&XG(internal_class_table));
 
-		zend_hash_init_ex(&XG(internal_function_table), 100, NULL, CG(function_table)->pDestructor, 1, 0);
-		zend_hash_init_ex(&XG(internal_class_table),    10,  NULL, CG(class_table)->pDestructor,    1, 0);
+#ifdef HAVE_XCACHE_CONSTANT
+		zend_hash_init_ex(&XG(internal_constant_table), 20,  NULL, (dtor_func_t) xc_zend_constant_dtor, 1, 0);
+#endif
+		zend_hash_init_ex(&XG(internal_function_table), 100, NULL, NULL, 1, 0);
+		zend_hash_init_ex(&XG(internal_class_table),    10,  NULL, NULL, 1, 0);
 
-		zend_hash_copy(&XG(internal_function_table), CG(function_table), (copy_ctor_func_t) function_add_ref, &tmp_func, sizeof(tmp_func));
-		zend_hash_copy(&XG(internal_class_table), CG(class_table), (copy_ctor_func_t) xc_zend_class_add_ref, &tmp_cest, sizeof(tmp_cest));
+#ifdef HAVE_XCACHE_CONSTANT
+		xc_copy_internal_zend_constants(&XG(internal_constant_table), EG(zend_constants));
+#endif
+		zend_hash_copy(&XG(internal_function_table), CG(function_table), NULL, &tmp_func, sizeof(tmp_func));
+		zend_hash_copy(&XG(internal_class_table), CG(class_table), NULL, &tmp_cest, sizeof(tmp_cest));
 
 		XG(internal_table_copied) = 1;
 	}
-	if (xc_php_hcache.size && !XG(php_holds)) {
+	if (xc_php_caches && !XG(php_holds)) {
 		XG(php_holds) = calloc(xc_php_hcache.size, sizeof(xc_stack_t));
 		for (i = 0; i < xc_php_hcache.size; i ++) {
 			xc_stack_init(&XG(php_holds[i]));
 		}
 	}
 
-	if (xc_var_hcache.size && !XG(var_holds)) {
+	if (xc_var_caches && !XG(var_holds)) {
 		XG(var_holds) = calloc(xc_var_hcache.size, sizeof(xc_stack_t));
 		for (i = 0; i < xc_var_hcache.size; i ++) {
 			xc_stack_init(&XG(var_holds[i]));
 		}
 	}
 
+#ifdef ZEND_ENGINE_2
+	zend_hash_init(&XG(gc_op_arrays), 32, NULL, xc_gc_op_array, 0);
+#endif
+
 #if PHP_API_VERSION <= 20041225
 	XG(request_time) = time(NULL);
 #else
@@ -1520,6 +1644,9 @@ static void xc_request_init(TSRMLS_D) /* {{{ */
 static void xc_request_shutdown(TSRMLS_D) /* {{{ */
 {
 	xc_entry_unholds(TSRMLS_C);
+#ifdef ZEND_ENGINE_2
+	zend_hash_destroy(&XG(gc_op_arrays));
+#endif
 	xc_gc_expires_php(TSRMLS_C);
 	xc_gc_expires_var(TSRMLS_C);
 	xc_gc_deletes(TSRMLS_C);
@@ -1538,6 +1665,9 @@ void xc_init_globals(zend_xcache_globals* xcache_globals TSRMLS_DC)
 {
 	memset(xcache_globals, 0, sizeof(zend_xcache_globals));
 
+#ifdef HAVE_XCACHE_CONSTANT
+	zend_hash_init_ex(&xcache_globals->internal_constant_table, 1, NULL, (dtor_func_t) xc_zend_constant_dtor, 1, 0);
+#endif
 	zend_hash_init_ex(&xcache_globals->internal_function_table, 1, NULL, NULL, 1, 0);
 	zend_hash_init_ex(&xcache_globals->internal_class_table,    1, NULL, NULL, 1, 0);
 }
@@ -1569,6 +1699,9 @@ void xc_shutdown_globals(zend_xcache_globals* xcache_globals TSRMLS_DC)
 	}
 
 	if (xcache_globals->internal_table_copied) {
+#ifdef HAVE_XCACHE_CONSTANT
+		zend_hash_destroy(&xcache_globals->internal_constant_table);
+#endif
 		zend_hash_destroy(&xcache_globals->internal_function_table);
 		zend_hash_destroy(&xcache_globals->internal_class_table);
 	}
@@ -1598,11 +1731,13 @@ static int xcache_admin_auth_check(TSRMLS_D) /* {{{ */
 	}
 
 	if (admin_user == NULL || admin_pass == NULL) {
-		php_error_docref(NULL TSRMLS_CC, E_ERROR, "xcache.admin.user and xcache.admin.pass is required");
+		php_error_docref(XCACHE_WIKI_URL "/InstallAdministration" TSRMLS_CC, E_ERROR,
+				"xcache.admin.user and/or xcache.admin.pass settings is not configured."
+				" Make sure you've modified the correct php ini file for your php used in webserver.");
 		zend_bailout();
 	}
 	if (strlen(admin_pass) != 32) {
-		php_error_docref(NULL TSRMLS_CC, E_ERROR, "unexpect %lu bytes of xcache.admin.pass, expected 32 bytes, the password after md5()", (unsigned long) strlen(admin_pass));
+		php_error_docref(NULL TSRMLS_CC, E_ERROR, "xcache.admin.pass is %lu chars unexpectedly, it is supposed to be the password after md5() which should be 32 chars", (unsigned long) strlen(admin_pass));
 		zend_bailout();
 	}
 
@@ -1645,13 +1780,28 @@ static int xcache_admin_auth_check(TSRMLS_D) /* {{{ */
 		}
 	}
 
-#define STR "WWW-authenticate: basic realm='XCache Administration'"
-	sapi_add_header_ex(STR, sizeof(STR) - 1, 1, 1 TSRMLS_CC);
-#undef STR
 #define STR "HTTP/1.0 401 Unauthorized"
 	sapi_add_header_ex(STR, sizeof(STR) - 1, 1, 1 TSRMLS_CC);
 #undef STR
-	ZEND_PUTS("XCache Auth Failed. User and Password is case sense\n");
+#define STR "WWW-authenticate: Basic Realm=\"XCache Administration\""
+	sapi_add_header_ex(STR, sizeof(STR) - 1, 1, 1 TSRMLS_CC);
+#undef STR
+#define STR "Content-type: text/html; charset=UTF-8"
+	sapi_add_header_ex(STR, sizeof(STR) - 1, 1, 1 TSRMLS_CC);
+#undef STR
+	ZEND_PUTS("<html>\n");
+	ZEND_PUTS("<head><title>XCache Authentication Failed</title></head>\n");
+	ZEND_PUTS("<body>\n");
+	ZEND_PUTS("<h1>XCache Authentication Failed</h1>\n");
+	ZEND_PUTS("<p>You're not authorized to access this page due to wrong username and/or password you typed.<br />The following check points is suggested:</p>\n");
+	ZEND_PUTS("<ul>\n");
+	ZEND_PUTS("<li>Be aware that `Username' and `Password' is case sense. Check capslock status led on your keyboard, and punch left/right Shift keys once for each</li>\n");
+	ZEND_PUTS("<li>Make sure the md5 password is generated correctly. You may use <a href=\"mkpassword.php\">mkpassword.php</a></li>\n");
+	ZEND_PUTS("<li>Reload browser cache by pressing F5 and/or Ctrl+F5, or simply clear browser cache after you've updated username/password in php ini.</li>\n");
+	ZEND_PUTS("</ul>\n");
+	ZEND_PUTS("Check <a href=\"" XCACHE_WIKI_URL "/InstallAdministration\">XCache wiki page</a> for more information.\n");
+	ZEND_PUTS("</body>\n");
+	ZEND_PUTS("</html>\n");
 
 	zend_bailout();
 	return 0;
@@ -1796,7 +1946,7 @@ static int xc_entry_init_key_var(xc_entry_t *xce, zval *name TSRMLS_DC) /* {{{ *
 			break;
 		default:
 #ifdef IS_UNICODE
-			convert_to_text(name);
+			convert_to_unicode(name);
 #else
 			convert_to_string(name);
 #endif
@@ -1854,7 +2004,7 @@ PHP_FUNCTION(xcache_get)
 		RETVAL_NULL();
 	} LEAVE_LOCK(xce.cache);
 	if (found) {
-		xce.cache->hits ++;
+		xc_cache_hit_dmz(xce.cache TSRMLS_CC);
 	}
 	else {
 		xce.cache->misses ++;
@@ -1935,7 +2085,7 @@ PHP_FUNCTION(xcache_isset)
 		RETVAL_FALSE;
 	} LEAVE_LOCK(xce.cache);
 	if (found) {
-		xce.cache->hits ++;
+		xc_cache_hit_dmz(xce.cache TSRMLS_CC);
 	}
 	else {
 		xce.cache->misses ++;
@@ -2003,14 +2153,10 @@ static inline void xc_var_inc_dec(int inc, INTERNAL_FUNCTION_PARAMETERS) /* {{{
 	ENTER_LOCK(xce.cache) {
 		stored_xce = xc_entry_find_dmz(&xce TSRMLS_CC);
 		if (stored_xce) {
-#ifdef DEBUG
-			fprintf(stderr, "incdec: gotxce %s\n", xce.name.str.val);
-#endif
+			TRACE("incdec: gotxce %s", xce.name.str.val);
 			/* timeout */
 			if (VAR_ENTRY_EXPIRED(stored_xce)) {
-#ifdef DEBUG
-				fprintf(stderr, "incdec: expired\n");
-#endif
+				TRACE("%s", "incdec: expired");
 				xc_entry_remove_dmz(stored_xce TSRMLS_CC);
 				stored_xce = NULL;
 			}
@@ -2021,9 +2167,7 @@ static inline void xc_var_inc_dec(int inc, INTERNAL_FUNCTION_PARAMETERS) /* {{{
 					zval *zv;
 					stored_xce->ctime = XG(request_time);
 					stored_xce->ttl   = xce.ttl;
-#ifdef DEBUG
-					fprintf(stderr, "incdec: islong\n");
-#endif
+					TRACE("%s", "incdec: islong");
 					value = Z_LVAL_P(stored_var->value);
 					value += (inc == 1 ? count : - count);
 					RETVAL_LONG(value);
@@ -2033,9 +2177,7 @@ static inline void xc_var_inc_dec(int inc, INTERNAL_FUNCTION_PARAMETERS) /* {{{
 					break; /* leave lock */
 				}
 				else {
-#ifdef DEBUG
-					fprintf(stderr, "incdec: notlong\n");
-#endif
+					TRACE("%s", "incdec: notlong");
 					xc_processor_restore_zval(&oldzval, stored_xce->data.var->value, stored_xce->have_references TSRMLS_CC);
 					convert_to_long(&oldzval);
 					value = Z_LVAL(oldzval);
@@ -2043,11 +2185,9 @@ static inline void xc_var_inc_dec(int inc, INTERNAL_FUNCTION_PARAMETERS) /* {{{
 				}
 			}
 		}
-#ifdef DEBUG
 		else {
-			fprintf(stderr, "incdec: %s not found\n", xce.name.str.val);
+			TRACE("incdec: %s not found", xce.name.str.val);
 		}
-#endif
 
 		value += (inc == 1 ? count : - count);
 		RETVAL_LONG(value);
@@ -2078,6 +2218,20 @@ PHP_FUNCTION(xcache_dec)
 	xc_var_inc_dec(-1, INTERNAL_FUNCTION_PARAM_PASSTHRU);
 }
 /* }}} */
+#ifdef HAVE_XCACHE_DPRINT
+/* {{{ proto bool  xcache_dprint(mixed value)
+   Prints variable (or value) internal struct (debug only) */
+PHP_FUNCTION(xcache_dprint)
+{
+	zval *value;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "z", &value) == FAILURE) {
+		return;
+	}
+	xc_dprint_zval(value, 0 TSRMLS_CC);
+}
+/* }}} */
+#endif
 /* {{{ proto string xcache_asm(string filename)
  */
 #ifdef HAVE_XCACHE_ASSEMBLER
@@ -2303,6 +2457,9 @@ static function_entry xcache_functions[] = /* {{{ */
 	PHP_FE(xcache_set,               NULL)
 	PHP_FE(xcache_isset,             NULL)
 	PHP_FE(xcache_unset,             NULL)
+#ifdef HAVE_XCACHE_DPRINT
+	PHP_FE(xcache_dprint,            NULL)
+#endif
 	{NULL, NULL,                     NULL}
 };
 /* }}} */
@@ -2440,6 +2597,21 @@ static PHP_MINFO_FUNCTION(xcache)
 	php_info_print_table_row(2, "Version", XCACHE_VERSION);
 	php_info_print_table_row(2, "Modules Built", XCACHE_MODULES);
 	php_info_print_table_row(2, "Readonly Protection", xc_readonly_protection ? "enabled" : "N/A");
+#ifdef ZEND_ENGINE_2_1
+	ptr = php_format_date("Y-m-d H:i:s", sizeof("Y-m-d H:i:s") - 1, xc_init_time, 1 TSRMLS_CC);
+	php_info_print_table_row(2, "Cache Init Time", ptr);
+	efree(ptr);
+#else
+	snprintf(buf, sizeof(buf), "%lu", (long unsigned) xc_init_time);
+	php_info_print_table_row(2, "Cache Init Time", buf);
+#endif
+
+#ifdef ZTS
+	snprintf(buf, sizeof(buf), "%lu.%lu", xc_init_instance_id, xc_init_instance_subid);
+#else
+	snprintf(buf, sizeof(buf), "%lu", xc_init_instance_id);
+#endif
+	php_info_print_table_row(2, "Cache Instance Id", buf);
 
 	if (xc_php_size) {
 		ptr = _php_math_number_format(xc_php_size, 0, '.', ',');
@@ -2484,17 +2656,15 @@ static PHP_MINFO_FUNCTION(xcache)
 /* {{{ extension startup */
 static void xc_zend_extension_register(zend_extension *new_extension, DL_HANDLE handle)
 {
-    zend_extension extension;
+	zend_extension extension;
 
-    extension = *new_extension;
-    extension.handle = handle;
+	extension = *new_extension;
+	extension.handle = handle;
 
-    zend_extension_dispatch_message(ZEND_EXTMSG_NEW_EXTENSION, &extension);
+	zend_extension_dispatch_message(ZEND_EXTMSG_NEW_EXTENSION, &extension);
 
-    zend_llist_prepend_element(&zend_extensions, &extension);
-#ifdef DEBUG
-	fprintf(stderr, "registered\n");
-#endif
+	zend_llist_prepend_element(&zend_extensions, &extension);
+	TRACE("%s", "registered");
 }
 
 static zend_llist_element *xc_llist_get_element_by_zend_extension(zend_llist *l, const char *extension_name)
@@ -2546,11 +2716,11 @@ static void xc_llist_unlink(zend_llist *l, zend_llist_element *element)
 
 static int xc_zend_extension_startup(zend_extension *extension)
 {
-    if (extension->startup) {
-        if (extension->startup(extension) != SUCCESS) {
+	if (extension->startup) {
+		if (extension->startup(extension) != SUCCESS) {
 			return FAILURE;
-        }
-    }
+		}
+	}
 	return SUCCESS;
 }
 /* }}} */
@@ -2638,16 +2808,6 @@ static PHP_MINIT_FUNCTION(xcache)
 #endif
 	REGISTER_INI_ENTRIES();
 
-	if (strcmp(sapi_module.name, "cli") == 0) {
-		if ((env = getenv("XCACHE_TEST")) != NULL) {
-			zend_alter_ini_entry("xcache.test", sizeof("xcache.test"), env, strlen(env) + 1, PHP_INI_SYSTEM, PHP_INI_STAGE_STARTUP);
-		}
-		if (!xc_test) {
-			/* disable cache for cli except for test */
-			xc_php_size = xc_var_size = 0;
-		}
-	}
-
 	xc_config_long(&xc_php_size,       "xcache.size",        "0");
 	xc_config_hash(&xc_php_hcache,     "xcache.count",       "1");
 	xc_config_hash(&xc_php_hentry,     "xcache.slots",      "8K");
@@ -2656,6 +2816,16 @@ static PHP_MINIT_FUNCTION(xcache)
 	xc_config_hash(&xc_var_hcache,     "xcache.var_count",   "1");
 	xc_config_hash(&xc_var_hentry,     "xcache.var_slots",  "8K");
 
+	if (strcmp(sapi_module.name, "cli") == 0) {
+		if ((env = getenv("XCACHE_TEST")) != NULL) {
+			zend_alter_ini_entry("xcache.test", sizeof("xcache.test"), env, strlen(env) + 1, PHP_INI_SYSTEM, PHP_INI_STAGE_STARTUP);
+		}
+		if (!xc_test) {
+			/* disable cache for cli except for testing */
+			xc_php_size = xc_var_size = 0;
+		}
+	}
+
 	if (xc_php_size <= 0) {
 		xc_php_size = xc_php_hcache.size = 0;
 	}
@@ -2676,6 +2846,15 @@ static PHP_MINIT_FUNCTION(xcache)
 			goto err_init;
 		}
 		xc_initized = 1;
+		xc_init_time = time(NULL);
+#ifdef PHP_WIN32
+		xc_init_instance_id = GetCurrentProcessId();
+#else
+		xc_init_instance_id = getpid();
+#endif
+#ifdef ZTS
+		xc_init_instance_subid = tsrm_thread_id();
+#endif
 	}
 
 #ifdef HAVE_XCACHE_COVERAGER
@@ -2846,7 +3025,7 @@ ZEND_DLEXPORT int xcache_zend_startup(zend_extension *extension) /* {{{ */
 		xc_llist_unlink(&zend_extensions, xc_llist_zend_extension);
 
 		ext = (zend_extension *) zend_llist_get_last_ex(&zend_extensions, &lpos);
-		assert(ext && ext != xc_llist_zend_extension);
+		assert(ext && ext != (zend_extension *) xc_llist_zend_extension->data);
 		xc_last_ext_startup = ext->startup;
 		ext->startup = xc_zend_startup_last;
 	}
