@@ -1,14 +1,16 @@
 #ifndef __XCACHE_H
 #define __XCACHE_H
 #define XCACHE_NAME       "XCache"
-#define XCACHE_VERSION    "1.2.2"
+#define XCACHE_VERSION    "1.3.0"
 #define XCACHE_AUTHOR     "mOo"
-#define XCACHE_COPYRIGHT  "Copyright (c) 2005-2007"
+#define XCACHE_COPYRIGHT  "Copyright (c) 2005-2009"
 #define XCACHE_URL        "http://xcache.lighttpd.net"
+#define XCACHE_WIKI_URL   XCACHE_URL "/wiki"
 
 #include <php.h>
 #include <zend_compile.h>
 #include <zend_API.h>
+#include <zend.h>
 #include "php_ini.h"
 #include "zend_hash.h"
 
@@ -37,6 +39,30 @@
 #	define ZESW(v1, v2) v2
 #endif
 
+#ifdef do_alloca_with_limit
+#	define my_do_alloca(size, use_heap) do_alloca_with_limit(size, use_heap)
+#	define my_free_alloca(size, use_heap) free_alloca_with_limit(size, use_heap)
+#elif defined(ALLOCA_FLAG)
+#	define my_do_alloca(size, use_heap) do_alloca(size, use_heap)
+#	define my_free_alloca(size, use_heap) free_alloca(size, use_heap)
+#else
+#	define my_do_alloca(size, use_heap) do_alloca(size)
+#	define my_free_alloca(size, use_heap) free_alloca(size)
+#	define ALLOCA_FLAG(x)
+#endif
+#ifndef Z_SET_ISREF
+#	define Z_SET_ISREF(z) (z).is_ref = 1
+#endif
+#ifndef Z_UNSET_ISREF
+#	define Z_UNSET_ISREF(z) (z).is_ref = 0
+#endif
+#ifndef Z_SET_REFCOUNT
+#	define Z_SET_REFCOUNT(z, rc) (z).refcount = rc
+#endif
+#ifndef IS_CONSTANT_TYPE_MASK
+#	define IS_CONSTANT_TYPE_MASK (~IS_CONSTANT_INDEX)
+#endif
+
 /* {{{ dirty fix for PHP 6 */
 #ifdef add_assoc_long_ex
 static inline void my_add_assoc_long_ex(zval *arg, char *key, uint key_len, long value)
@@ -118,9 +144,15 @@ typedef char *zstr;
 #	define zend_u_hash_add(ht, type, arKey, nKeyLength, pData, nDataSize, pDest) \
  	   zend_hash_add(ht, arKey, nKeyLength, pData, nDataSize, pDest)
 
+#	define zend_u_hash_quick_add(ht, type, arKey, nKeyLength, h, pData, nDataSize, pDest) \
+ 	   zend_hash_quick_add(ht, arKey, nKeyLength, h, pData, nDataSize, pDest)
+
 #	define zend_u_hash_update(ht, type, arKey, nKeyLength, pData, nDataSize, pDest) \
  	   zend_hash_update(ht, arKey, nKeyLength, pData, nDataSize, pDest)
 
+#	define zend_u_hash_quick_update(ht, type, arKey, nKeyLength, h, pData, nDataSize, pDest) \
+ 	   zend_hash_quick_update(ht, arKey, nKeyLength, h, pData, nDataSize, pDest)
+
 #	define zend_u_hash_find(ht, type, arKey, nKeyLength, pData) \
  	   zend_hash_find(ht, arKey, nKeyLength, pData)
 
@@ -191,6 +223,13 @@ typedef struct {
 
 	time_t     last_gc_deletes;
 	time_t     last_gc_expires;
+
+	time_t     hits_by_hour_cur_time;
+	zend_uint  hits_by_hour_cur_slot;
+	zend_ulong hits_by_hour[24];
+	time_t     hits_by_second_cur_time;
+	zend_uint  hits_by_second_cur_slot;
+	zend_ulong hits_by_second[5];
 } xc_cache_t;
 /* }}} */
 /* {{{ xc_classinfo_t */
@@ -200,8 +239,11 @@ typedef struct {
 #endif
 	zstr      key;
 	zend_uint key_size;
+	ulong     h;
 	xc_cest_t cest;
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 	int       oplineno;
+#endif
 } xc_classinfo_t;
 /* }}} */
 #ifdef HAVE_XCACHE_CONSTANT
@@ -212,6 +254,7 @@ typedef struct {
 #endif
 	zstr      key;
 	zend_uint key_size;
+	ulong     h;
 	zend_constant constant;
 } xc_constinfo_t;
 /* }}} */
@@ -223,6 +266,7 @@ typedef struct {
 #endif
 	zstr      key;
 	zend_uint key_size;
+	ulong     h;
 	zend_function func;
 } xc_funcinfo_t;
 /* }}} */
@@ -234,6 +278,7 @@ typedef struct {
 #endif
 	zstr       key;
 	zend_uint  key_len;
+	ulong      h;
 } xc_autoglobal_t;
 /* }}} */
 #endif
@@ -259,7 +304,9 @@ typedef struct {
 
 	zend_uint classinfo_cnt;
 	xc_classinfo_t *classinfos;
+#ifndef ZEND_COMPILE_DELAYED_BINDING
 	zend_bool have_early_binding;
+#endif
 
 #ifdef ZEND_ENGINE_2_1
 	zend_uint autoglobal_cnt;
@@ -308,5 +355,6 @@ extern zend_module_entry xcache_module_entry;
 int xc_is_rw(const void *p);
 int xc_is_ro(const void *p);
 int xc_is_shm(const void *p);
+void xc_gc_add_op_array(zend_op_array *op_array TSRMLS_DC);
 
 #endif /* __XCACHE_H */
